import itertools
from typing import *

import pandas


def to_squareform(pairwise_values: Dict[Tuple[str,str], Any], default:Any = None)->pandas.DataFrame:
	"""
		Converts a dictionary of pariwise values (ex. a representation of pairwise distances)
		into a square matrix, where the keys in the labels are converted
		to columns and index labels.
		Ex.
		data = {
			('a', 'b'): 'ab',
			('b', 'c'): 'bc',
			('a', 'c'): 'ac'
		}
		matrix:
			'a'	'b'	'c'
		'a'	None	'ab'	'ac'
		'b'	'ab'	None	'bc'
		'c'	'ac'	'bc'	None
	"""

	keys = sorted(set(itertools.chain.from_iterable(pairwise_values)))
	_square_map = dict()
	for left in keys:
		series = dict()
		for right in keys:
			value = pairwise_values.get((left, right))
			if value is None:
				value = pairwise_values.get((right, left), default)
			series[right] = value
		_square_map[left] = series
	return pandas.DataFrame(_square_map)



from typing import *
from pathlib import Path
import numpy

def calculate_area_of_polygon(x:numpy.ndarray, y:numpy.ndarray):
	return 0.5 * numpy.abs(numpy.dot(x, numpy.roll(y, 1)) - numpy.dot(y, numpy.roll(x, 1)))

def main():
	pass


if __name__ == '__main__':
	main()



import datetime
from dataclasses import dataclass

import pendulum
import pytest

from infotools.timetools import Duration


@pytest.fixture
def duration():
	data = datetime.timedelta(days = 12, seconds = 1245, microseconds = 123)
	return data


@dataclass
class Generic:
	days: int
	seconds: int
	microseconds: int


@pytest.mark.parametrize(
	"data",
	[
		"P12DT20M45.000123S",
		Generic(12, 1245, 123),
		pendulum.Duration(12, 1245, 123),
		datetime.timedelta(12, 1245, 123),
		(12, 1245, 123),
		{'days': 12, 'seconds': 1245, 'microseconds': 123},
	]
)
def test_parse_duration(data, duration):
	result = Duration(data)
	assert result == duration


@pytest.mark.parametrize(
	"string, expected",
	[("P00Y00W01DT00H30M34S", datetime.timedelta(days = 1, hours = 0, minutes = 30, seconds = 34))]
)
def test_parse_string_duration(string, expected):
	result = Duration.from_iso(string)
	assert result == Duration.from_timedelta(expected)


@pytest.mark.parametrize(
	"value, expected",
	[
		("00:23:17", 1397),
		("23:17", 1397)
	]
)
def test_parse_duration_variable(value, expected: int):
	# `expected` should be the expected number of seconds.
	result = Duration(value)
	assert result.total_seconds() == expected


@pytest.mark.parametrize(
	"value, expected",
	[
		("00:23:17", 1397),
		("23:17", 1397)
	]
)
def test_parse_duration_from_string(value, expected):
	result = Duration.from_string(value)
	assert result.total_seconds() == expected


def test_duration_attributes():
	# PT8H37M8.070428S 31028.070428 31028
	total_seconds = 31029.070428
	obj = Duration(seconds = total_seconds)
	assert obj.to_iso() == "PT08H37M09S"
	assert obj.days == 0
	assert obj.hours == 8
	assert obj.minutes == 37
	assert obj.seconds == 31029  # Done by pendulum to maintain compatibility with datetime.timedelta.
	assert obj.remaining_seconds == 9
	assert obj.microseconds == 70428


@pytest.mark.parametrize(
	"seconds, expected",
	[
		(10, "00:00:10.00"),
		(31029.070428, "08:37:09.07"),
		(3601, "01:00:01.00")
	]
)
def test_to_standard(seconds, expected):
	duration = Duration(seconds = seconds)
	assert duration.to_standard() == expected


@pytest.mark.parametrize(
	"seconds, expected",
	[
		(100, datetime.timedelta(seconds = 100)),
		(24 * 3600 + 7234, datetime.timedelta(seconds = 24 * 3600 + 7234))
	]
)
def test_to_timedelta(seconds, expected):
	duration = Duration(seconds = seconds)
	result = duration.to_timedelta()

	assert isinstance(result, datetime.timedelta)
	assert result == expected


def test_duration_repr(duration):
	duration = Duration(duration)
	expected = {'days': 12, 'seconds': 1245, 'microseconds': 123}
	assert duration.to_dict() == expected

	expected = {'years': 0, 'weeks': 1, 'days': 5, 'hours': 0, 'minutes': 20, 'seconds': 45, 'microseconds': 123}
	assert duration.tolongdict() == expected

	expected = "P00Y01W05DT00H20M45S"
	assert duration.to_iso() == expected
	expected = "P00Y01W05DT00H20M45S"
	assert duration.to_iso(compact = False) == expected

	td = datetime.timedelta(days = 12, seconds = 1245, microseconds = 123)
	assert isinstance(td, datetime.timedelta)
	assert duration.to_timedelta() == td
	# data = datetime.timedelta(days = 12, seconds = 1245, microseconds = 123)
	total_days = 12 + (1245.123 / 86400)
	total_years_expected = total_days / 365
	assert duration.total_years() == pytest.approx(total_years_expected)


@pytest.mark.parametrize(
	"seconds, expected",
	[
		(10, {'days': 0, 'seconds': 10, 'microseconds': 0}),
		(31029.070428, {'days': 0, 'seconds': 31029, 'microseconds': 70428}),
		(86401.0123, {'days': 1, 'seconds': 1, 'microseconds': 12300})
	]
)
def test_to_dict(seconds, expected):
	duration = Duration(seconds = seconds)
	assert duration.to_dict() == expected


@pytest.mark.parametrize(
	"seconds, expected",
	[
		(10, 'PT10S'),
		(86401, "P01DT01S")
	]

)
def test_to_iso_compact(seconds, expected):
	result = Duration(seconds = seconds).to_iso(compact = True)
	assert result == expected


@pytest.mark.parametrize(
	"seconds, expected",
	[
		(10, 'PT00H00M10S'),
		(86401.0123, "P00Y00W01DT00H00M01S")
	]

)
def test_to_iso_full(seconds, expected):
	result = Duration(seconds = seconds).to_iso(compact = False)
	assert result == expected


@pytest.mark.parametrize(
	"seconds, expected",
	[
		(10, 'PT00H00M10S'),
		(86401.0123, "P00Y00W01DT00H00M01.0123S"),
		(3601, "PT01H00M01S"),
		(0, "PT00H00M00S")
	]

)
def test_to_iso_medium(seconds, expected):
	result = Duration(seconds = seconds).to_iso(compact = False, include_microseconds = True)
	assert result == expected


@pytest.mark.parametrize(
	"value, expected",
	[
		("00:00:10.00", 10),
		("00:00:11", 11),
		("00:12", 12),
		("55:13", 3313),
		("4:55:14", 17714),
		("4:55:14.5", 17714.5),

	]
)
def test_from_standard(value, expected):
	result = Duration.from_string(value)
	assert result.total_seconds() == expected



from pathlib import Path
import numpy
import pytest
from loguru import logger

from infotools import filetools
import datetime


@pytest.mark.parametrize(
	"filename, expected",
	[
		("file1.mp4", ('video', 'mp4')),
		("abc.aac", ("audio", "aac"))
	]
)
def test_get_mimetype(filename, expected):
	assert filetools.get_mimetype(filename) == expected


def test_checkdir(tmp_path):
	folder = Path(__file__).parent / "new"
	result = filetools.checkdir(folder)

	logger.debug(f"Folder: {folder}, {type(folder)}, {folder.exists()}")
	logger.debug(f"result: {result}, {type(result)}, {result.exists()}")
	logger.debug(f"{result == folder}")
	assert result == folder
	assert result.exists()


@pytest.mark.parametrize(
	"obj, expected",
	[
		(Path(__file__), __file__),  # Convert Path to str
		(numpy.linspace(0, 10, 21), [0, 0.5, 1, 1.5, 2, 2.5, 3, 3.5, 4, 4.5, 5, 5.5, 6, 6.5, 7, 7.5, 8, 8.5, 9, 9.5, 10]),
		(numpy.float64(10), 10.0),
		(numpy.float64(5.3), 5.3),
		(datetime.date(2022, 7, 13), "2022-07-13"),
		(datetime.datetime(2022, 9, 13, 4, 9, 39), "2022-09-13T04:09:39"),
		(datetime.timedelta(seconds = 123), f"0:02:03")
	]
)
def test_to_json(obj, expected):
	result = filetools.to_json(obj)
	assert result == expected
	assert type(result) == type(expected)



import math

import pandas

from infotools import datatools


def test_to_squareform():
	data = {
		('a', 'b'): 'ab',
		('b', 'c'): 'bc',
		('a', 'c'): 'ac'
	}

	expected = {
		'a': [math.nan, 'ab', 'ac'],
		'b': ['ab', math.nan, 'bc'],
		'c': ['ac', 'bc', math.nan]
	}
	expected = pandas.DataFrame(expected, index = ['a', 'b', 'c'])
	result = datatools.to_squareform(data)

	pandas.testing.assert_frame_equal(result, expected)



"""
	Suite of tests for numbertools
"""

import pytest

from infotools import numbertools


@pytest.fixture
def decimal() -> numbertools.DecimalScale:
	return numbertools.DecimalScale()


def test_decimalscale_get_unit_magnitude(decimal):
	result = decimal.get_unit_magnitude()
	assert result.multiplier == 1


@pytest.mark.parametrize(
	"value, expected_multiplier",
	[
		(0.00013, 1E-6),
		(0.1010, 1E-3),
		(0, 1),
		(1, 1),
		(10, 1),
		(12_123_456, 1E6)
	]
)
def test_decimalscale_get_magnitude_from_value(decimal, value, expected_multiplier):
	result = decimal.get_magnitude_from_value(value)
	assert result.multiplier == expected_multiplier


@pytest.mark.parametrize(
	"prefix, expected_multiplier",
	[
		("μ", 1E-6),
		("micro", 1E-6),
		("m", 1E-3),
		("milli", 1E-3),
		('', 1),
		('M', 1E6),
		('mega', 1E6),
		('P', 1E15),
		('peta', 1E15)
	]
)
def test_decimalscale_get_magnitude_from_prefix(decimal, prefix, expected_multiplier):
	result = decimal.get_magnitude_from_prefix(prefix)
	assert result.multiplier == expected_multiplier


@pytest.mark.parametrize(
	"alias, expected_multiplier",
	[
		('unit', 1),
		("millionths", 1E-6),
		("trillion", 1E12),
	]
)
def test_decimalscale_get_magnitude_from_alias(decimal, alias, expected_multiplier):
	result = decimal.get_magnitude_from_alias(alias)
	assert result.multiplier == expected_multiplier
@pytest.mark.parametrize(
	"value, expected_multiplier",
	[
		('u', 1E-6),
		("μ", 1E-6),
		('millionths', 1E-6),
		(0.000123, 1E-6),
		("micro", 1E-6),
		("m", 1E-3),
		("milli", 1E-3),
		("thousandths", 1E-3),
		(0.123, 1E-3),
		('', 1),
		('K', 1E3),
		('kilo', 1E3),
		('thousand', 1E3),
		('Thousand', 1E3),
		(453999.999, 1E3),
		('M', 1E6),
		('mega', 1E6),
		(123_456_789, 1E6),
		('P', 1E15),
		('peta', 1E15)
	]
)
def test_decimalscale_get_magnitude(decimal, value, expected_multiplier):
	result = decimal.get_magnitude(value)
	assert result.multiplier == expected_multiplier

@pytest.mark.parametrize(
	"source, target, expected",
	[
		('unit', 'K', 1E-3),
		('unit', 'f', 1E15),
		('micro', 'K', 1E-9),
		('B', 'T', 1E-3),
		('T', 'unit', 1E12),
		('T', 'B', 1E3),
		(123_456_789_111_222, 'unit', 1E12)
	]
)
def test_get_multiplier(decimal, source, target, expected):
	result = decimal.get_multiplier(source, target)
	assert result == pytest.approx(expected)


@pytest.mark.parametrize(
	"value, target, expected",
	[
		(0, 'f', 0),
		(0, 'K', 0),
		(1, 'f', 1E15),
		(1, 'B', 1E-9),
		(123_456_789_111_222, 'unit', 123_456_789_111_222),
		(123_456_789_111_222, 'T', 123.456789111222),
		(123_456_789_111_222, 'B', 123456.789111222),
		(123_456_789_111_222, 'm', 123_456_789_111_222_000),
		(-500_000_000_000.0,'T', -0.500),
		(1234.123,"unit", 1234.123),
		(1234.123,"", 1234.123)
	]
)
def test_decimalscale_convert_value(decimal, value, target, expected):
	result = decimal.convert(value, target)
	assert pytest.approx(result) == expected


@pytest.fixture
def binary() -> numbertools.BinaryScale:
	return numbertools.BinaryScale()


@pytest.mark.parametrize("value,expected",
	[
		(123.456, True),
		('123.456', True),
		('abc', False),
		('12.345.678', False),
		("1E6", True),
		([123, "456.f", "789.0"], [True, False, True])
	])
def test_is_number(value, expected):
	assert expected == numbertools.is_number(value)


@pytest.mark.parametrize(
	"number,precision,base,expected",
	[
		(1234.123, 6, None, '1.234123K'),
		(12.5E-6, 1, None, '12.5μ'),
		(111_222_333_444_555, 12, None, '111.222333444555T'),
		(-1234.123, 6, None, '-1.234123K'),
		(-500_000_000_000.0, 2, None, "-500.00B"),
		(-500_100_000_000.0, 2, None, "-500.10B"),
		(-500_000_000_000.0, 0, None, "-500B"),
		(0.0, 0, None, "0"),
		(0.0, 2, None, "0.00"),
		(1234.123, 3, "unit", '1234.123'),
		(111_222_333_444_555, 2, 'B', '111222.33B'),
		(-500_000_000_000.0, 3, 'T', "-0.500T"),
		(-500_000_000_000.0, 2, 'T', "-0.50T")
	]
)
def test_human_readable_1(number, precision, base, expected):
	assert numbertools.human_readable(number, precision = precision, base = base) == expected

@pytest.mark.parametrize(
	"number, precision, base, system, expected",
	[
		(1_000_000, 2, None, 'decimal', '1.00M'),
		(1_000_000, 2, None, 'binary', '976.56KiB'),
		(1_000_000_000, 1, None, 'decimal', '1.0B'),
		(1_000_000_000, 1, None, 'binary', '953.7MiB')
	]
)
def test_human_readable_2(number, precision, base, system, expected):
	assert numbertools.human_readable(number, precision = precision, base = base, system = system) == expected
@pytest.mark.parametrize(
	"value,expected",
	[
		("5.4", 5.4),
		('asfkjnlmlqwe', None),
		('7.8/10 ', 0.78)
	]
)
def test_convert_string_to_number(value, expected):
	result = numbertools.to_number(value, default = None)

	assert result == expected


@pytest.mark.parametrize(
	"value,expected",
	[
		("5.4", 5.4),
		('asfkjnlmlqwe', None),
		('7.8/10 ', 0.78),
		(['7.8/10', '99'], [0.78, 99])
	]
)
def test_to_number(value, expected):
	result = numbertools.to_number(value, None)

	assert result == expected


@pytest.mark.parametrize(
	"value",
	[
		"kilo", "micro", "milli", "femto"
	]
)
def test_get_magnitude_from_prefix(value):
	result = numbertools.DecimalScale().get_magnitude_from_prefix(value)
	assert result.prefix == value


@pytest.mark.parametrize(
	"system", [numbertools.DecimalScale(), numbertools.BinaryScale()]
)
def test_get_unit_magnitude(system):
	assert system.get_unit_magnitude() == system.get_magnitude_from_alias('unit')


@pytest.mark.parametrize(
	"value,expected",
	[
		(1234, 'kilo'),
		(1E-4, 'micro'),
		(0.123, 'milli'),
		(0.0, "")
	]
)
def test_get_magnitude_from_value(decimal, value, expected):
	result = decimal.get_magnitude_from_value(value)
	assert result.prefix == expected


@pytest.mark.parametrize(
	"value,expected",
	[
		('billion', 'giga'),
		('Millions', 'mega')
	]
)
def test_get_magnitude_from_alias(decimal, value, expected):
	result = decimal.get_magnitude_from_alias(value)

	assert result.prefix == expected



import math
from dataclasses import dataclass, field
from typing import *
from loguru import logger
from fuzzywuzzy import process

NumberType = Union[int, float]


@dataclass
class Magnitude:
	""" Provides an easy method of checking the magnitude of numbers."""
	prefix: str
	suffix: str
	multiplier: NumberType
	alias: List[str] = field(default_factory = list)  # Alternative methods of referring to this multiplier.

	def __str__(self)->str:
		label = f"Magnitude({self.prefix=}, {self.suffix=}, {self.multiplier=:E}, {self.alias=})"
		return label
	@staticmethod
	def _get_other(other):
		""" Returns the value in `other` that the dunder methods need to compare.
		"""
		if hasattr(other, 'multiplier'):
			return other.multiplier
		else:
			return other

	def __float__(self) -> float:
		return float(self.multiplier)

	def __post_init__(self):
		self.alias.append(self.prefix)

	def __mul__(self, other) -> float:
		return self._get_other(other) * self.multiplier

	def __rmul__(self, other) -> float:
		return self.__mul__(other)

	def __ge__(self, other):
		other = self._get_other(other)
		return self.multiplier >= other

	def __gt__(self, other) -> bool:
		other = self._get_other(other)
		return self.multiplier > other

	def __lt__(self, other):
		other = self._get_other(other)
		return self.multiplier < other

	def __le__(self, other):
		other = self._get_other(other)
		return self.multiplier <= other

	def __eq__(self, other):
		other = self._get_other(other)
		return self.multiplier == other

	def is_match(self, value: str) -> bool:
		""" Returns True if the passed string corresponds to this scale."""
		selected_scale = False
		if self.prefix == value or self.suffix == value:
			selected_scale = True
		else:
			scale_alias, score = process.extractOne(value.lower(), self.alias)
			if score > 90:
				selected_scale = True
		return selected_scale


class AbstractScale:
	def __init__(self):
		self.system: List[Magnitude] = []

	@staticmethod
	def is_null(value) -> bool:
		""" Checks if a value represents a null value."""
		try:
			result = value is None or math.isnan(float(value))
		except (TypeError, ValueError):
			result = True

		return result

	def get_unit_magnitude(self):
		raise NotImplementedError

	def get_magnitude_from_value(self, value: SupportsAbs) -> Magnitude:
		value = abs(value)

		if value == 0.0 or self.is_null(value):
			return self.get_unit_magnitude()

		for _scale in self.system[::-1]:

			if value >= _scale.multiplier:
				magnitude = _scale
				break
		else:
			message = f"'{value}' does not have a defined base."
			raise ValueError(message)

		return magnitude

	def get_magnitude_from_prefix(self, prefix: str) -> Optional[Magnitude]:
		"""
			Retrieves the `Magnitude` object where `Magnitude.prefix` or `Magnitude.suffix` equals `prefix`
		"""
		try:
			candidates = [i for i in self.system if (i.prefix == prefix or i.suffix == prefix)]
			return candidates[0]
		except IndexError:
			return None

	def get_magnitude_from_alias(self, alias: str) -> Optional[Magnitude]:
		for element in self.system:
			if not element.alias:
				# Don't bother with empty aliases.
				continue
			candidate, score = process.extractOne(alias.lower(), element.alias)
			if score > 90:
				return element
		# Added to make it clear the method should return `None`
		return None

	def get_magnitude(self, value:Union[NumberType, str])->Optional[Magnitude]:
		"""
			Returns a `Magnitude` object where one attribute matches `value`.
		"""
		if isinstance(value, str):
			magnitude = self.get_magnitude_from_prefix(value)
			if magnitude is None:
				magnitude = self.get_magnitude_from_alias(value)
		elif isinstance(value, Magnitude):
			magnitude = value
		else:
			magnitude = self.get_magnitude_from_value(value)
		return magnitude

	def get_multiplier(self, source:Union[str, NumberType,Magnitude], target:str)->NumberType:
		"""
			Returns the multipler to convert from the scale `source` to `target`.
			Parameters
			----------
			source, target: Union[str, NumberType,Magnitude]
				Identifies the `Magnitude` to use.
		"""
		source_magnitude = self.get_magnitude(source)
		target_magnitude = self.get_magnitude(target)

		multiplier = source_magnitude.multiplier / target_magnitude.multiplier
		return multiplier


	def convert_scale(self, value:NumberType,  target:str, base:str = None):
		"""
			Converts a number to a target scale.
			Parameters
			----------
			value: NumberType
				The value to convert.
			target: str
				The prefix or alias of the magnitude to convert to. Ex. 'T', 'femto'

			Returns
			-------
			NumberType
		"""
	def convert(self, value: NumberType, target: str, base:Optional[str] = 'unit') -> NumberType:
		"""
			Converts a number to a target scale.
			Parameters
			----------
			value: NumberType
				The value to convert.
			target: str
				The prefix or alias of the magnitude to convert to. Ex. 'T', 'femto'
			base: Optional[str] = 'unit'
				The current base of `value`. Defaults to `unit` so by default the function simply converts the number as given
				into the target base. If given, the method will convert `value` from the source base to the target base.
			Returns
			-------
			NumberType
		"""
		multiplier = self.get_multiplier(base, target)
		return value * multiplier


class DecimalScale(AbstractScale):
	def __init__(self):
		super().__init__()
		self.base = 10
		self.system = [
			Magnitude('atto', 'a', 1E-18),
			Magnitude('femto', 'f', 1E-15),
			Magnitude('pico', 'p', 1E-12),
			Magnitude('nano', 'n', 1E-9),
			Magnitude('micro', "μ", 1E-6, ["u", 'millionths']),
			Magnitude('milli', 'm', 1E-3, ['thousandths']),
			Magnitude('', '', 1E0, ['unit', 'one']),
			Magnitude('kilo', 'K', 1E3, ['thousand']),
			Magnitude('mega', 'M', 1E6, ['million']),
			Magnitude('giga', 'B', 1E9, ['billion']),
			Magnitude('tera', 'T', 1E12, ['trillion']),
			Magnitude('peta', 'P', 1E15, ['quadrillion']),
			Magnitude('exa', 'E', 1E18, ['quintillion'])
		]


	def get_unit_magnitude(self) -> Magnitude:
		return self.system[6]


class BinaryScale(AbstractScale):
	def __init__(self):
		super().__init__()
		self.base = 1024
		self.system = [
			Magnitude('', '', self.base ** 0, ['unit', '']),
			Magnitude('kibi', 'KiB', self.base ** 1, ['thousand']),
			Magnitude('mebi', 'MiB', self.base ** 2, ['million']),
			Magnitude('gibi', 'GiB', self.base ** 3, ['billion']),
			Magnitude('tebi', 'TiB', self.base ** 4, ['trillion']),
			Magnitude('pebi', 'PiB', self.base ** 5, ['quadrillion']),
			Magnitude('exbi', 'EiB', self.base ** 6, ['quintillion']),
			Magnitude('zebi', 'ZiB', self.base ** 7, []),
			Magnitude('yobi', 'YiB', self.base ** 8, [])
		]

	def get_unit_magnitude(self):
		return self.system[0]


if __name__ == "__main__":
	scale = BinaryScale()
	value = 1_000_000_000
	print(scale.get_magnitude_from_value(value))



from typing import *

def calculate_luminance(color: str) -> float:
	# 0.299 * color.R + 0.587 * color.G + 0.114 * color.B
	red = int(color[1:3], 16)
	green = int(color[3:5], 16)
	blue = int(color[5:], 16)

	lum = (.299 * red) + (.587 * green) + (.114 * blue)
	return lum / 255 

def hex_to_rgb(string, as_sRGB: bool = False) -> Tuple[int, int, int]:
	red = string[1:3]
	green = string[3:5]
	blue = string[5:7]

	red = int(red, 16)
	green = int(green, 16)
	blue = int(blue, 16)

	if as_sRGB:
		red = red / 55
		green = green / 255
		blue = blue / 255

	return red, green, blue


def rgb_to_hex(value: Union[str, Tuple[int, int, int]]) -> str:
	""" Converts the color specified in the tiff file to hex format. The color is usually given as either as a string ('255,0,255') or tuple ((255, 0, 255))"""
	if isinstance(value, str):
		value = [int(i) for i in value.split(',')]

	if all(i <= 1 for i in value):  # Float RGB format
		color = mcolors.rgb2hex(value).upper()
	else:
		value_red, value_green, value_blue, *extra = value
		color = f"#{value_red:>02X}{value_green:>02X}{value_blue:>02X}"
	return color


def convert_to_luminance_factor(value: float) -> float:
	""" Converts an sRGB value to whatever unit luminance uses. """
	if value < 0.03928:
		value = value / 12.92
	else:
		value = ((value + 0.055) / 1.055) ** 2.4
	return value

def calculate_luminance(rgb: Tuple[int, int, int]) -> float:
	red, green, blue = rgb
	red = convert_to_luminance_factor(red)
	green = convert_to_luminance_factor(green)
	blue = convert_to_luminance_factor(blue)

	luminance = (0.2126 * red) + (0.7152 * green) + (0.0722 * blue)
	if luminance == 0:
		luminance = 1
	return luminance

def calculate_contrast(color_1: str, color_2: str) -> float:
	rgb_1 = hex_to_rgb(color_1, as_sRGB = True)
	rgb_2 = hex_to_rgb(color_2, as_sRGB = True)
	# the relative luminance of the lighter colour (L1) is divided through the relative luminance of the darker colour (L2):
	l_1 = calculate_luminance(rgb_1)
	l_2 = calculate_luminance(rgb_2)
	# Check if the darker/lighter color was reversed
	if l_2 > l_1:
		l_1, l_2 = l_2, l_1

	ratio = (l_1 + 0.05) / (l_2 / 0.05)

	return ratio


"""
	A version of timetools built on Pendulum. Pendulum has a number of great features, but suffers from the
	same drawbacks as other Date/time modules when creating an object from another object or uncommon format.
	Pendulum also does not offer convienience methods for some datetime representations (ex. ISO durations).
	Ex. pandas.Timestamp is not compatible with pendulum.datetime.
"""

import datetime
import re
from dataclasses import dataclass
from typing import *

import pendulum


@dataclass
class TimedeltaInformation:
	"""Helps keep track of data being passed around."""
	days: int
	seconds: int
	microseconds: int

	def to_dict(self):
		""" Converts `self` to a dict"""
		data = {
			'days':         self.days,
			'seconds':      self.seconds,
			'microseconds': self.microseconds
		}
		return data


class Duration(pendulum.Duration):
	"""
		A drop-in replacement for datetime and Pendulum. Contains a number or
		useful methods for time timedelta representations.
	"""

	def __new__(cls, value = None, **kwargs):
		"""
			Keyword Arguments
			-----------------
			All arguments are optional and default to 0.
			Arguments may be integers or floats, and may be positive or negative.
				days:         number; default 0
				seconds:      number; default 0
				microseconds: number; default 0
				milliseconds: number; default 0
				minutes:      number; default 0
				hours:        number; default 0
				weeks:        number; default 0
		"""
		if value is not None:
			return cls.parse(value)

		return super().__new__(cls, **kwargs)

	# self.duration = pendulum.duration(**duration_keys)

	def __str__(self):
		return self.to_iso()

	def __repr__(self):
		repr_string = "Duration('{}')".format(self.to_iso())
		return repr_string

	@classmethod
	def parse(cls, value: Any) -> 'Duration':
		"""
			Attempts to parse/coerce any object into a timedelta.
			The given element will be parsed by pendulum.parse, when possible.
		Parameters
		----------
		value

		Returns
		-------

		"""
		if isinstance(value, str):
			# result = pendulum.parse(value)
			# result = cls.from_object(result)
			result = cls.from_string(value)
		elif isinstance(value, dict):
			result = cls.from_keys(value)

		elif isinstance(value, (list, tuple)):
			result = cls.from_tuple(value)
		else:
			result = cls.from_object(value)

		return result

	@classmethod
	def from_dict(cls, **keys) -> 'Duration':
		""" initializes a `Duration` object from a dictionary using `**` notation."""
		return cls(**keys)

	@classmethod
	def from_keys(cls, keys: Dict[str, int]) -> 'Duration':
		""" initializes a `Duration` object from a dictionary."""
		return cls(**keys)

	@classmethod
	def from_standard(cls, value: str) -> 'Duration':
		""" Initializes a `Duration` object from a string formatted as 'HH:MM:SS.SS'"""
		times = value.split(':')
		if len(times) == 1:
			hour = minute = 0
			second = float(times[0])
		elif len(times) == 2:
			hour = 0
			minute, second = map(float, times)  # use float to preserve milliseconds
		else:
			hour, minute, second = map(float, times)
		return cls(hours = hour, minutes = minute, seconds = second)

	@classmethod
	def from_string(cls, string: str) -> 'Duration':
		"""
			Parses a string. Defaults to pendulum.parse
		Parameters
		----------
		string: str

		Returns
		-------
		Duration
		"""
		if ':' in string:
			return cls.from_standard(string)
		else:
			result = pendulum.parse(string)
			return cls.from_object(result)

	@classmethod
	def from_object(cls, obj: Any) -> 'Duration':
		"""
			Attempts to convert a generic object to a timedelta Duration.
		Parameters
		----------
		obj: Any

		Returns
		-------
		Duration
		"""
		if hasattr(obj, 'as_timedelta'):
			obj = obj.as_timedelta()  # Converts pendulum.Duration to a timedelta.
		if isinstance(obj, datetime.timedelta) or hasattr(obj, 'total_seconds'):
			seconds = obj.total_seconds()
			days = 0
			microseconds = 0
		elif isinstance(obj, pendulum.Duration):
			seconds = obj.in_seconds()
			days = 0
			microseconds = 0
		else:
			try:
				days = obj.days
				seconds = obj.seconds
				microseconds = obj.microseconds
			except AttributeError:
				message = "Object '{}' with type '{}' cannot be coerced to a time duration.".format(obj, type(obj))
				raise AttributeError(message)

		result = {
			'days':         days,
			'seconds':      seconds,
			'microseconds': microseconds
		}
		return cls.from_dict(**result)

	@classmethod
	def from_timedelta(cls, obj: datetime.timedelta) -> 'Duration':
		""" Wrapper around .from_object that allows explicit initialization from a timedelta."""
		return cls.from_object(obj)

	@classmethod
	def from_tuple(cls, value: Tuple) -> 'Duration':
		"""
			Attempts to construct a timedelta form an iterable.
		Parameters
		----------
		value: Tuple[int,int,int]
			The value must consist of three parts: days, seconds, microseconds.

		Returns
		-------
		Duration
		"""
		if len(value) != 3:
			message = f"The value passed to Duration.from_tuple must contain exactly 3 values (recieved {value})."
			raise ValueError(message)
		keys = ['days', 'seconds', 'microseconds']
		timedelta_keys = dict(zip(keys, value))
		return cls.from_dict(**timedelta_keys)

	@classmethod
	def from_iso(cls, value: str) -> int:
		""" Generates a Duration object from an iso-formatted string."""
		pattern = "[\d]{2}[YWDHMS]"

		multipliers = {
			'Y': 365 * 24 * 3600,
			'W': 7 * 24 * 3600,
			'D': 24 * 3600,
			'H': 3600,
			'M': 60,
			'S': 1
		}

		matches = re.findall(pattern, value)
		total_seconds = 0
		for match in matches:
			multiplier = multipliers[match[-1]]
			value = int(match[:2]) * multiplier
			total_seconds += value
		return cls(seconds = total_seconds)

	def to_dict(self) -> Dict[str, int]:
		""" Returns a dictionary that can be used to instantiate another timedelta or Duration object. """
		result = {
			'days':         self.days,
			'seconds':      self.seconds,
			'microseconds': self.microseconds
		}
		return result

	def tolongdict(self) -> Dict[str, int]:
		""" Returns a dictionary with more human readable date and time keys. """

		is_negative = self.total_seconds() < 0
		if is_negative:
			original = abs(self)
			original = {
				'days':         original.days,
				'seconds':      original.seconds,
				'microseconds': original.microseconds
			}
		else:
			original = self.to_dict()
		longdict = dict()
		# Get date values
		days = original['days']
		longdict['years'], days = divmod(days, 365)
		longdict['weeks'], longdict['days'] = divmod(days, 7)

		# Get time values
		seconds = original['seconds']
		longdict['hours'], seconds = divmod(seconds, 3600)
		longdict['minutes'], longdict['seconds'] = divmod(seconds, 60)
		# longdict['seconds'] += original['microseconds'] / 1000000
		longdict['microseconds'] = original['microseconds']

		# Since timedelta objects subtract positive numbers from the largest unit for negative timestamps, need to convert back,

		return longdict

	def to_iso(self, compact: bool = False, include_microseconds: bool = False) -> str:
		""" Converts the timedelta to an ISO Duration string. By default,
			weeks are used instead of months, so the original duration string
			used to create to Duration object may differ (but will be equivilant to)
			the output of this method.
			Parameters
			----------
				compact: bool; default False
					Whether to omit empty fields.
		"""
		is_negative = self.total_seconds() < 0
		values = self.tolongdict()

		if not compact:
			string = "P{years:>02}Y{weeks:>02}W{days:>02}DT{hours:>02}H{minutes:>02}M{seconds:>02}S"
			result = string.format(**values)
			if include_microseconds and values['microseconds'] > 0:
				seconds = values['seconds'] + (values['microseconds'] / 1E6)

				second_string = str(seconds) + 'S'
				if len(second_string.split('.')[0]) == 1:
					second_string = '0' + second_string

				result = result.split('M')[0] + 'M' + second_string

			if self.total_seconds() < 86400:
				result = 'PT' + result.split('T')[-1]
			return result

		suffix_map = [
			('years', 'Y'), ('months', 'M'), ('weeks', 'W'), ('days', 'D'),
			('hours', 'H'), ('minutes', 'M'), ('seconds', 'S')
		]
		suffix_map = dict(suffix_map)

		large_keys = ['years', 'weeks', 'days']
		small_keys = ['hours', 'minutes', 'seconds']
		# Modify the "seconds" value so it has two digits before the decimal point.
		if values["seconds"] < 10:
			values["seconds"] = "0" + str(values['seconds'])

		large_values = "P" + "".join(["{0:>02}{1}".format(values[key], suffix_map[key]) for key in large_keys if values[key] != 0])
		small_values = "T" + "".join(["{0:>02}{1}".format(values[key], suffix_map[key]) for key in small_keys if values[key] != 0])

		isostring = large_values + small_values

		if compact:
			if isostring == 'PT':  # Duration of 0 seconds
				isostring = 'PT00S'

		if is_negative:
			isostring = '-' + isostring
		return isostring

	def to_timedelta(self) -> datetime.timedelta:
		""" Returns a timedelta equivilant to `self`"""
		return self.as_timedelta()

	def to_standard(self) -> str:
		"""
			Returns the duration formatted as HH:MM:SS.SS. Currently only designed for timedeltas less than a day.
		"""
		hours = self.hours
		minutes = self.minutes
		seconds = self.remaining_seconds + (self.microseconds / 1E6)

		result = f"{hours:>02}:{minutes:>02}:{seconds:>05.2f}"
		return result

	def total_years(self) -> float:
		"""
			Returns the total number of years contained in the duration as a float.
		Returns
		-------

		"""
		return self.total_days() / 365

	def to_json(self) -> str:
		""" Returns a json-safe representation of `self`"""
		return str(self)

	def to_yaml(self) -> str:
		""" Returns a yaml representation of `self`"""
		return self.to_json()



"""
	A simple timer for tracking how long snippets of code take to run.
"""
import statistics
import time
from typing import Callable, Dict, Tuple, Union, Optional

try:
	from infotools.timetools import Duration
	from infotools import numbertools
except ModuleNotFoundError:
	from .. import numbertools
	from ._duration import Duration

Number = Union[int, float]


class Timer:
	""" A class with convienient timing methods.
		.timeit -> Times a number of loops
		.timeFunction -> benchmarks a function
		.benchmark -> benchmarks an external process and returns
			both the average execution time and standard deviation.

		Example
		-------
		```
		with Timer():
			result = method_to_time(*args)


		```
	
	"""

	def __init__(self, label: Optional[str] = None):
		self.start_time = time.clock()
		self.end_time = 0.0
		self.label = label

	def __str__(self):
		return self.duration.to_iso()

	def __enter__(self):
		self.start = time.time()

	def __exit__(self, exc_type, exc_val, exc_tb):
		duration = time.time() - self.start
		print(f"Finished {self.label} in {duration:.2f} seconds.", flush = True)

	@property
	def duration(self) -> Duration:
		return Duration(time.clock() - self.start_time)

	def is_over(self, limit: Number = 10.0) -> bool:
		""" Checks if more time has elapsed than the supplied limit.
			Parameters
			----------
				limit: int, float; default 10.0
					The time limit, in seconds. 
					If more than <limit> seconds have passed, returns True
			Returns
			----------
				duration : bool
		"""
		result = self.duration.total_seconds() >= limit
		return result

	def togo(self, done: int, total: int) -> Duration:
		""" Calculates the remaining time until a loop is finished
			executing
			Parameters
			----------
				done: int
					The number of loops that have already passed
				total: int
					The total number of loops expected
			Returns
			----------
				remaining : float, string
		"""
		if done == 0: done += 1
		perloop = self.duration / done
		remaining = (total - done) * perloop
		return Duration(remaining)

	def reset(self) -> None:
		self.__init__()

	def benchmark(self, loops: int = 1) -> Dict[str, Number]:
		""" Returns a dictionary with information on the loop timing.
		
			Returns
			-------
				result: dict<>
					* 'duration': int
						The total time that has passed.
					* 'perLoop': float
						The average number of seconds passed for every loop.
					* 'loops': int
						Number of loops passed to the function.
		"""
		duration = self.duration
		per_loop = duration / loops

		result = {
			'duration': Duration(seconds = duration),
			'perLoop':  per_loop,
			'loops':    loops
		}
		return result

	def time_function(self, func: Callable, *args, **kwargs) -> Tuple[str, str]:
		""" Benchmarks a function. args and kwargs are passed on to the function.
			Prints a message of the form 'a ± b per loop [c, d] where:
				* 'a': average time for each loop to execute. 
				* 'b': standard deviation for each loop. 
				* 'c': The fastest time any given loop completed. 
				* 'd': The slowest time any given loop took to complete.
			Parameters
			----------
			func: callable
				The function to benchmark.
			* 'loops': int; default 10
				The number of times to run the function during the benchmark.
			Returns
			-------
				result: tuple -> float, float
					The average and standard deviation of the loop execution times.
		"""
		if 'loops' in kwargs:
			loops = kwargs.pop('loops')
		else:
			loops = 100
		_results = list()

		for _ in range(loops):
			self.reset()
			func(*args, **kwargs)
		minimum = min(_results)
		maximum = max(_results)

		minimum = numbertools.human_readable(minimum)
		maximum = numbertools.human_readable(maximum)
		avg = numbertools.human_readable(statistics.mean(_results))
		std = numbertools.human_readable(statistics.stdev(_results))
		print("{}s ± {}s per loop [{} loops][{}s, {}s]".format(avg, std, loops, minimum, maximum))
		return avg, std

	def timeit(self, loops: int = 1, label: str = None):
		""" Calculates the time for a loop(s) to execute. Resets the timer.
			Parameters
			----------
				loops: int
					The total number of loops that were executed.
				label: string
					Added to the timeit message if provided.
			Return
			----------
				duration : string
					a string of the form 'a per loop (b for c loops)'
					* 'a': The average execution time for each loop 
					* 'b': The total execution time
					* 'c': The number of loops executed.
		"""
		benchmark = self.benchmark(loops)
		duration = benchmark['duration']
		per_loop = numbertools.human_readable(benchmark['perLoop'])

		if label is None:
			message = ""
		else:
			message = label + ': '

		message = message + "{0}s per loop ({1:.2f}s for {2:n} loop(s)) ".format(per_loop, duration, loops)

		self.reset()
		return message



from pathlib import Path
from typing import Dict, Union

import pandas


def read_table(file_name: Union[str, Path], **kwargs):
	""" Reads the table and returns a dataframe. This is basically just a short script that lets
		users import data without having to worry about filetype.
	"""
	file_name = Path(file_name)
	extension = file_name.suffix
	default_args = {
		'.csv': {'delimiter': ','},
		'.tsv': {'delimiter': '\t'}
	}

	# arguments = self._cleanArguments(extension, arguments)
	file_name = str(file_name.absolute())
	if extension in {'.xls', '.xlsx', '.xlsm'}:  # .xlsm is not a typo.

		df = pandas.read_excel(file_name, **kwargs)
	elif extension in {'.csv', '.tsv', '.fsv', '.txt'}:
		arguments = {**default_args.get(extension), **kwargs}
		if 'sheetname' in arguments: arguments.pop('sheetname')
		df = pandas.read_table(file_name, **arguments)
	elif extension == '.pkl':
		df = pandas.read_pickle(file_name)
	else:
		raise NameError("{} does not have a valid extension!".format(file_name))
	return df


def to_spreadsheet(tables: Dict[str, pandas.DataFrame], filename: Path, include_index: bool = True) -> Path:
	"""
		Saves the table as an Excel spreadsheet, where multiple tables can be given..
	Parameters
	----------
	tables: Dict[str,pandas.DataFrame]
		A mapping of sheet names to dataframes.

	filename: str, pathlib.Path
		The output file.

	Returns
	-------
	Path: The output filename
	"""
	#writer = pandas.ExcelWriter(str(filename))
	# python 3.5 or 3.6 made all dicts ordered by default, so the sheets will be ordered in the same order they were defined in `tables`
	with pandas.ExcelWriter(str(filename)) as writer:
		for sheet_label, table in tables.items():
			if table is None:
				continue
			table.to_excel(writer, sheet_name = sheet_label, index = include_index)

	# Need to use the xlsx library to change some features of the spreadsheet.

	return filename

def to_csv(table:pandas.DataFrame, filename:Path, precision:int = 2):
	"""
		A wrapper around csv/pandas that removes insignificant digits from the values in numerical columns.
		Parameters
		----------
		filename:Path
			Filename to the output table
	"""

	for column_label in table.columns:
		column = table[column_label]
		is_numeric = ""
		is_float = ""

		if is_numeric:
			minimum_value = column.min()
			maximum_value = column.max()

			value = float(f"{value:.2f}")




from typing import *

import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from infotools import numbertools
import random

class Palette:
	def __init__(self, colormap: str = "Blues", step:int = 10):
		self.breakpoints = [i * step for i in range(10)]
		self.colors = seaborn.color_palette(colormap, len(self.breakpoints) + 1)

		self.colormap_bins = {cutoff: color for cutoff, color in zip(self.breakpoints, self.colors)}

	def get_color(self, value: float) -> str:
		for cutoff, color in self.colormap_bins.items():
			if value <= cutoff:
				break
		else:
			color = '#FF0000'
		return color

	def add_legend(self, ax: plt.Axes, colormap: Dict[str, str] = None, **kwargs) -> plt.Axes:
		if colormap is None:
			colormap = self.colormap_bins
		ax = add_legend(ax, colormap,**kwargs)
		return ax



def add_legend(ax: plt.Axes, colormap: Dict[str, str], keep_order:bool = False,  **kwargs) -> plt.Axes:

	items = colormap.items() if keep_order else sorted(colormap.items())

	patches = list()
	for label, color in items:
		patch = mpatches.Patch(color = color, label = label)
		patches.append(patch)

	ax.legend(handles = patches, **kwargs)
	return ax


def get_random_color(lower: int = 50, upper: int = 250) -> str:
	red = random.randint(lower, upper)
	green = random.randint(lower, upper)
	blue = random.randint(lower, upper)
	return f"#{red:>02X}{green:>02X}{blue:>02X}"


def hr_labels(ax: plt.Axes, which: Literal['x', 'y', 'xy', 'both'] = 'y') -> plt.Axes:
	""" Converts numerical values into a human-friendly format.
		Parameters
		----------
		ax:plt.Axes
			The ax object that needs to be modified.
		which: Literal['x', 'y', 'xy', 'both']; default 'y'
			Specifies which axis to modify.
		Returns
		-------
		plt.Axes
			The modified ax object.
	"""
	if which in {'x', 'xy', 'both'}:
		x_ticks = ax.xaxis.get_majorticklocs()
		x_labels = [numbertools.human_readable(i) for i in x_ticks]
		ax.set_xticklabels(x_labels)

	if which in {'y', 'xy', 'both'}:
		y_ticks = ax.yaxis.get_majorticklocs()
		y_labels = [numbertools.human_readable(i) for i in y_ticks]
		ax.set_yticklabels(y_labels)

	return ax



from setuptools import setup
"""
python setup.py sdist bdist_wheel
twine upload -repository pypi dist/*
"""

import infotools
"""
if infotools.DEBUG:
	message = f"The module is still in debug mode!"
	raise ValueError(message)
"""
setup(
	name = 'infotools',
	version = '2025.05.06',
	packages = ['infotools', 'infotools.timetools', 'infotools.numbertools'],
	url = 'https://github.com/Kokitis/infotools',
	license = 'MIT',
	author = 'proginoskes',
	author_email = 'chrisdeitrick1@gmail.com',
	description = 'A collection of tools to make common tasks simpler.',
	classifiers = [
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
	], install_requires = ['pendulum', 'fuzzywuzzy', 'loguru', 'psutil', 'pandas', 'tqdm'],
	tests_requires = ['pytest']
)



"""
	Convienient methods for converting between numbers and strings and number representations.
"""

import math
from numbers import Number
from typing import Any, Iterable, List, Union, Optional, Literal
import numpy
from loguru import logger

# from ._scale import scale
try:
	from . import _scale
except ImportError:
	import _scale
default_scale = _scale.DecimalScale()

available_systems = {
	'decimal': _scale.DecimalScale(),
	'binary': _scale.BinaryScale()
}

NumberType = Union[int, float]


def _is_null(value) -> bool:
	if value is None or not isinstance(value, (int, float)):
		return True
	return False


def human_readable(value: NumberType, precision: int = 2, base: Optional[str] = None, system:Literal['decimal', 'binary'] = 'decimal') -> str:
	""" Converts a number into a more easily-read string.
		Ex. 101,000,000,000,000 -> '101T'

		Parameters
		----------
		value: NumberType
			Any number or list of numbers. If a list is given, all numbers
			will be asigned the same suffix as the lowest number.
		precision: int; default 2
			The number of decimal places to show.
		base: Optional[str]
			Converts `value` to the given base. `None` skips conversion. Assumes the values are from metric units.
		system: Literal['decimal', 'binary']; default 'decimal'
			The scale with which to choose the number prefix/suffix from.
		Returns
		-------
		str, list<str>
			The reformatted number.
	"""
	if system is None:
		system = 'decimal'
	if system not in available_systems:
		message = f"'{system}' is not an available scalling system. Choose from one of {list(available_systems.keys())}"
		raise ValueError(message)
	else:
		current_scale = available_systems[system]
	template = '{0:.' + str(int(precision)) + 'f}{1}'
	if base is None:
		magnitude = current_scale.get_magnitude_from_value(value)
		human_readable_number = value / magnitude.multiplier
		string = template.format(human_readable_number, magnitude.suffix)
	else:
		logger.debug(f"{value=}")
		human_readable_number = current_scale.convert(value, base)
		logger.debug(f"{human_readable_number=}")
		string = template.format(human_readable_number, base if base not in {'', 'unit'} else "")

	return string


def is_number(value: Union[Any, Iterable[Any]]) -> Union[bool, List[bool]]:
	"""Tests if the value is a number.

		Examples
		--------
		'abc'->False
		123.123 -> True
		'123.123' -> True

	"""
	if isinstance(value, (list, tuple)):
		return [is_number(i) for i in value]
	if isinstance(value, str):
		try:
			float(value)
			value_is_number = True
		except ValueError:
			value_is_number = False
	else:
		value_is_number = isinstance(value, Number)

	return value_is_number


def _convert_string_to_number(value: str, default = math.nan) -> float:
	if '/' in value:
		left, right = value.split('/')
		left = _convert_string_to_number(left)
		right = _convert_string_to_number(right)
		return left / right
	else:
		value = value.replace(',', '')  # Remove thousands separator.
		value = value.strip()
		try:
			value = float(value)
		except ValueError:
			value = default
		return value


def to_number(value: Union[Any, Iterable[Any]], default: Any = math.nan) -> Union[NumberType, List[NumberType]]:
	""" Attempts to convert the passed object to a number.
		Returns
		-------
			value: Scalar
				* list,tuple,set -> list of Number
				* int,float -> int, float
				* str -> int, float
				* generic -> float if float() works, else math.nan
	"""
	if isinstance(value, str):
		return _convert_string_to_number(value, default)

	if isinstance(value, (list, tuple, set)):
		return [to_number(i, default) for i in value]

	try:
		converted_number = float(value)
	except (ValueError, TypeError):
		converted_number = default

	if not _is_null(converted_number) and math.floor(converted_number) == converted_number:
		converted_number = int(converted_number)

	return converted_number


def to_decimal(value, base, lower = False, readable = False):
	""" Converts any number to base 10
		Parameters
		----------
			value: string
				The number to convert
			base: int
				The current base of the number
			readable: bool; default False
				Whether to omit similar symbols. Currently defunct
		Returns
		----------
			number: int
	"""
	Z = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'
	if lower:
		Z = Z.lower()
	if base < 37:
		return int(value, base)
	else:
		Z = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'
	digits = value[::-1]
	number = sum(Z.find(digit) * (base ** index) for index, digit in enumerate(digits))

	return number


def from_decimal(value, base, readable = False):
	""" Converts a number in Base 10 to another Base
		Parameters
		----------
			value: int
				The number (in base 10) to convert
			base: int
				The base to convert the number to
			readable: bool, default False
				Omits similar symbols. Currently non-working
		Returns
		----------
			number : string
	"""
	if base < 37:
		Z = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'
	else:
		Z = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'
	if base > len(Z):
		print("to_base({0}, {1})".format(value, base))
		raise ValueError("base must be >= 2 and <= {0}".format(len(Z)))
	_pythonic_bases = {2, 8, 16}
	if base in _pythonic_bases:
		if base == 2:
			number = '{0:b}'.format(value)
		elif base == 8:
			number = '{0:o}'.format(value)
		elif base == 16:
			number = '{0:X}'.format(value)
		return number

	bases = list()
	index = 0
	maximum = 1
	while maximum <= value:
		index += 1
		bases.append(maximum)
		maximum = pow(base, index)
	bases = bases[::-1]

	number = str()
	for ibase in bases:
		mod, value = divmod(value, ibase)
		number += Z[mod]
	return number


def convert_base(value, from_base, to_base):
	return from_decimal(to_decimal(value, from_base), to_base)


def normalize_values(values: numpy.ndarray) -> numpy.ndarray:
	"""
		Maps the values form the input into the domain [0,1]
	"""
	return (values - numpy.min(values)) / (numpy.max(values) - numpy.min(values))


def main():
	parameters = [
		(111_222_333_444_555, 6, 'B', '111222.33B'),
		(-500_000_000_000.0, 6, 'T', '-0.500T'),
		(1234.123, 6, "unit", '1234.123')
	]
	precision = 6

	for value, precision, base, expected in parameters:
		result = human_readable(value, precision, base)
		print(result)


if __name__ == "__main__":
	main()



"""
	A version of timetools built on Pendulum. Pendulum has a number of great features, but suffers from the
	same drawbacks as other Date/time modules when creating an object from another object or uncommon format.
	Pendulum also does not offer convienience methods for some datetime representations (ex. ISO durations).
	Ex. pandas.Timestamp is not compatible with pendulum.datetime.
"""

import datetime
import re
from typing import *

import pendulum
from loguru import logger

STuple = Tuple[int, ...]
TTuple = Tuple[int, int, int]


def _attempt_to_get_attribute(obj: Any, key: str, default = 0):
	try:
		attribute = getattr(obj, key)
	except AttributeError:
		attribute = default
	return attribute


class Timestamp(pendulum.DateTime):
	def __new__(cls, *args, **kwargs):
		if len(args) == 1:
			value = args[0]
		elif len(args) > 1:
			return cls.from_values(*args)
		else:
			value = None
		if value is not None:
			return cls.parse(value)
		result = super().__new__(cls, **kwargs)
		return result

	def __repr__(self) -> str:
		""" This changes what repr() returns for Timestamp objects so they are shown with ISO timestamps.
		ex. "Timestamp(2013, 10, 23, 0, 0, 0)" -> "Timestamp('2013-10-23T00:00:00')"
		"""
		iso_string = self.to_iso()
		result = f"Timestamp('{iso_string}')"
		return result

	def __eq__(self, other):
		return self.year == other.year and self.month == other.month and self.day == other.day and self.hour == other.hour and self.minute == other.minute and self.second == other.second

	def __float__(self) -> float:
		""" Converts the timestamp to a floating point representation.
			Ex. float(Timestamp('2018-06-31')) -> 2018.5
		"""
		return self.year + (self.day_of_year / 365)

	@classmethod
	def parse(cls, value: Any) -> 'Timestamp':
		if isinstance(value, str):
			result = cls.from_string(value)
		elif isinstance(value, (list, tuple)):
			result = cls.from_tuple(value)
		elif isinstance(value, dict):
			result = cls.from_keys(value)
		else:
			result = cls.from_object(value)
		return result

	@classmethod
	def from_dict(cls, **kwargs) -> 'Timestamp':
		result = cls(**kwargs)
		return result

	@classmethod
	def from_keys(cls, keys: Dict[str, int]) -> 'Timestamp':
		return cls.from_dict(**keys)

	@classmethod
	def from_tuple(cls, value: Union[STuple, TTuple]) -> 'Timestamp':
		if len(value) == 3:
			year, month, day = value
			hour, minute, second = 0, 0, 0
			other = []
		else:
			year, month, day, hour, minute, second, *other = value

		data = {
			'year':   year,
			'month':  month,
			'day':    day,
			'hour':   hour,
			'minute': minute,
			'second': second
		}
		if len(other) > 0:
			data['microsecond'] = other[0]
		else:
			data['microsecond'] = 0
		return cls.from_dict(**data)

	@classmethod
	def from_object(cls, obj: Any) -> 'Timestamp':
		"""
			Attempts to create a pendulum.DateTime object from another datetime object from a
			different module.
		Parameters
		----------
		obj: Any
			Should have .year, .month, and .day methods, but may also have .hour, .minute, .hour, .tz attributes.

		Returns
		-------
		Timestamp
		"""
		year = obj.year
		month = obj.month
		day = obj.day

		hour = _attempt_to_get_attribute(obj, 'hour', 0)
		minute = _attempt_to_get_attribute(obj, 'minute', 0)
		second = _attempt_to_get_attribute(obj, 'second', 0)
		microsecond = _attempt_to_get_attribute(obj, 'microsecond', 0)

		result = cls.from_values(year, month, day, hour, minute, second, microsecond)

		return result

	@classmethod
	def from_american_date(cls, value: str) -> pendulum.DateTime:
		"""
			Parses a date formatted as DD/MM/YY(YY), as is common in the US.

		Parameters
		----------
		value:str

		Returns
		-------
		pendulum.DateTime
		"""
		if ' ' in value:
			dates, times = value.split(' ')
		elif 'T' in value:
			dates, times = value.split('T')
		else:
			dates = value
			times = ""

		month, day, year = list(map(int, dates.split('/')))
		# Need to fix the year vlue if it's only twp digits
		if year < 1900:
			if year > 40:  # "Close to the midpoint of the century."
				year += 1900
			else:
				year += 2000
		if times:
			hour, minute, second, *_ = list(map(int, times.split(':')))
		else:
			hour, minute, second = 0, 0, 0

		keys = {
			'year':   year,
			'month':  month,
			'day':    day,
			'hour':   hour,
			'minute': minute,
			'second': second
		}

		return cls.from_dict(**keys)

	@classmethod
	def from_verbal_date(cls, value: str) -> Optional["Timestamp"]:
		# 17 Dec 2012
		verbal_regex_month_first = "(?P<month>[a-z]+)\s(?P<day>[\d]+)[\s,]+(?P<year>[\d]{4})"
		verbal_regex_day_first = "(?P<day>[\d]+)[\s,]+(?P<month>[a-z]+)[.]?\s(?P<year>[\d]{4})"
		value = value.lower()

		short_months = ["jan", "feb", "mar", "apr", "may", "jun", "jul", "aug", "sep", "oct", "nov", "dec"]
		long_months = ["january", "february", "april", "may", "june", "july", "august", "september", "october",
			"november", "december"]

		match = re.search(verbal_regex_month_first, value)
		if not match:
			match = re.search(verbal_regex_day_first, value)
		if match:
			groupdict = match.groupdict()
			year = int(groupdict['year'])
			month = groupdict['month']
			day = int(groupdict['day'])
			if len(month) == 3:
				month = short_months.index(month) + 1
			else:
				month = long_months.index(month) + 1

			data = {
				'day':   day,
				'month': month,
				'year':  year
			}
			return cls.from_dict(**data)

	@classmethod
	def from_string(cls, value: str) -> 'Timestamp':

		try:
			obj = pendulum.parse(value)

		except ValueError:
			try:
				obj = cls.from_american_date(value)
			except ValueError:
				obj = cls.from_verbal_date(value)

		return cls.from_object(obj)

	@classmethod
	def from_regex(cls, value: str, regex: str = None):
		""" Uses regular expressoin to parse the input timestamp.
			Each regular expression should return a dictionary with the keys 'year', 'month', and 'day'
		"""

		regexes = [
			# '13 Sep. 2005', '1 Dec. 2021', "20 Apr. 2022"
			"(?P<day>[\d]{1,2})\s(?P<month>(jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec))[.]?\s(?P<year>[\d]{4})"
		]
		# Ignore capitalization
		value = value.lower()
		for regex in regexes:
			match = re.search(regex, value)
			if match: break
		else:
			match = None

		if match:
			match = match.groupdict()
			match['day'] = int(match['day'])
			match['year'] = int(match['year'])
			match['month'] = parse_month_string(match['month'])
			obj = cls.from_dict(**match)
		else:
			logger.debug(f"Regex failed: {value=}\t{match=}")
			obj = None
		return obj

	@classmethod
	def from_values(cls, year, month, day, hour = 0, minute = 0, second = 0, microsecond = 0,
		timezone = None) -> 'Timestamp':
		result = dict(
			year = year,
			month = month,
			day = day,
			hour = hour,
			minute = minute,
			second = second,
			microsecond = microsecond
		)
		return cls.from_dict(**result)

	def to_iso(self) -> str:
		return self.to_iso8601_string()

	def to_datetime(self) -> datetime.datetime:
		return datetime.datetime(
			year = self.year, month = self.month, day = self.day,
			hour = self.hour, minute = self.minute, second = self.second, microsecond = self.microsecond
		)


def parse_month_string(string: str) -> int:
	""" Attempts to convert the input string into an integer representing the month """
	months = ['jan', 'feb', 'mar', 'apr', 'may', 'jun', 'jul', 'aug', 'sep', 'oct', 'nov', 'dec']
	alias = {name: index for index, name in enumerate(months, start = 1)}

	return alias.get(string.lower())


def main():
	value = '13 Sep. 2005'
	ts = Timestamp.from_string(value)
	print(ts)


if __name__ == "__main__":
	main()



import hashlib
import mimetypes
import os
from pathlib import Path
from typing import Tuple, Union, Set
import string
import re

mimetypes.add_type('audio/aac', '.aac')

from loguru import logger

Pathlike = Union[str, Path]


def get_type_name(t) -> str:
	pattern = "[\w]+[.][\w]+"
	match = re.search(pattern, str(t))
	if match:
		match = match.group(0)
	return match


def get_mimetype(filename: Pathlike) -> Tuple[str, str]:
	""" Wrapper to get the mimetype of a given file. Returns `None` if the mimetype cannot be determined.
		Returns
		-------
		mimetype, filetype
			Ex. ('text', 'plain'), ('video', 'mp4')

	"""
	# Cast to Path so that we can use Path methods
	filename = Path(filename)
	# TODO: Include `folder` as a valid mimetype/filetype?
	mtype = mimetypes.guess_type(str(filename))
	mtype, *_ = mtype
	if mtype:
		type_mime = tuple(mtype.split('/'))  # Cast to tuple for consistency
	else:
		logger.warning(f"Could not determine the mimetype of {filename}: {mtype}")
		type_mime = 'unknown', filename.suffix
	return type_mime


def memory_usage(show = True, units = 'MB', label: str = ""):
	""" Gets the current memory usage
		Returns
		----------
			if show is False
			memory: int
				The total number of bytes being used by the current process
	"""
	import psutil
	process = psutil.Process(os.getpid())
	usage = process.memory_info().rss
	if show:
		if units == 'MB':
			value = usage / 1024 ** 2
		else:
			value = usage

		# print("Current memory usage: {0:.2f}{1}".format(value, units), flush = True)
		print(f"Current memory usage: {value:.2f}")
	return usage


def checkdir(path: Pathlike) -> Path:
	""" Creates a folder if it doesn't already exist.
		Parameters
		----------
			path: Path
				Path to a folder.
		Returns
		-------
		Path: The path that was checked.
	"""
	path = Path(path)
	# if path.is_dir() and not path.exists():
	if path.is_dir() and not path.exists():
		path.mkdir()
	return path


def copyfile(source: Path, target: Path) -> Path:
	target.write_bytes(source.read_bytes())
	return target


def generate_md5(filename: Union[str, Path], blocksize: int = 2 ** 20) -> str:
	""" Generates the md5sum of a file. Does
		not require a lot of memory.
		Parameters
		----------
			filename: string
				The file to generate the md5sum for.
			blocksize: int; default 2**20
				The amount of memory to use when
				generating the md5sum string.
		Returns
		-------
			md5sum: string
				The md5sum string.
	"""
	m = hashlib.md5()
	with open(str(filename), "rb") as f:
		while True:
			buf = f.read(blocksize)
			if not buf: break
			m.update(buf)
	return m.hexdigest()


def get_allowed_characters() -> Set[str]:
	allowed_characters = set(string.ascii_letters + string.digits)
	allowed_characters = allowed_characters | {'.', ',', '&', '(', ')', '+', '-', '[', ']', '_'}
	to_keep = {'Ñ', 'Ö', 'ß', 'á', 'ã', 'ä', 'ç', 'è', 'é', 'ê', 'ñ', 'ó', 'ö', 'ü', 'ę', 'Μ', 'Φ', 'К', 'М', 'Т', 'а', 'в', 'е', 'и', 'к', 'н', 'р',
		'т', 'у', 'х', 'я', '/', ' '}
	allowed_characters = allowed_characters | to_keep

	return allowed_characters


def sanitize_text(text: str, replacement: str = "") -> str:
	allowed_characteres = get_allowed_characters()
	characters = set(text)
	characters_to_remove = characters - allowed_characteres
	characters_to_replace = {' -': "", '/': "-", '&amp': "&"}
	for char in characters_to_remove:
		text = text.replace(char, replacement)

	for character_to_replace, replacement in characters_to_replace.items():
		text = text.replace(character_to_replace, replacement)

	# Remove all extra whitespace
	# pattern = "[\s]+"
	# text = re.sub(pattern, " ", text)
	return text


def to_json(obj, filename: Path = None) -> str:
	""" Tries to convert datatypes to json-usable versions. Ex numpy.ndarray -> list(). """
	import json
	import numpy
	import datetime

	# Here's a map of which python types need to be converted to json types.
	type_map = {
		int:   {numpy.integer},
		float: {numpy.floating},
		list:  {numpy.ndarray},
		str:   {Path}
	}

	# Also implement a way of detecting whether `obj` has a method to convert it to json.
	possible_methods = ['to_json', 'save_json', 'json', 'isoformat']

	class NpEncoder(json.JSONEncoder):

		def convert_numpy_object(self, obj):
			""" Tries to convert an object from numpy into a compatible object. """
			obj = self.convert_numpy_numeric(obj)
			if isinstance(obj, numpy.ndarray):
				return obj.tolist()
			return obj

		@staticmethod
		def convert_numpy_numeric(obj):
			object_type = type(obj)
			for key_type, candidates in type_map.items():
				if object_type in candidates:
					return key_type(obj)
			return obj

		def call_conversion_method(self, obj):
			for method in possible_methods:
				if hasattr(obj, method):
					attribute = getattr(obj, method)
					return attribute()

		def default(self, obj):
			type_name = get_type_name(type(obj))
			if type_name.startswith('numpy'):
				return self.convert_numpy_object(obj)

			if isinstance(obj, Path):
				return str(obj)
			if isinstance(obj, (datetime.datetime, datetime.date)):
				return obj.isoformat()
			if isinstance(obj, datetime.timedelta):
				return str(obj)
			# Now try to detect custom json implementations.
			if hasattr(obj, 'to_json'):
				return obj.to_json()
			return super(NpEncoder, self).default(obj)

	content = json.dumps(obj, cls = NpEncoder)
	if filename:
		filename.write_text(content)

	content = json.loads(content)

	return content


if __name__ == "__main__":
	pass



from pathlib import Path

from infotools import tabletools

folder_data = Path(__file__).parent / "data"


def test_read_table():
	""" Basically just tests if the function crashes when any table from the `data` folder is given to it.
	"""
	# ignore .txt file for now since they're not guaranteed to be tables.
	allowed_extensions = {'.csv', '.tsv', 'xls', 'xlsx', 'xlsm'}
	filenames = [i for i in folder_data.iterdir() if i.suffix in allowed_extensions]

	for filename in filenames:
		tabletools.read_table(filename)



"""
	A suite of tests to ensure timetools.Timestamp is operating properly.
"""
import datetime

import pandas
import pendulum
import pytest

from infotools import timetools


@pytest.fixture
def timestamp():
	key = "2019-05-06 00:14:26.246155"
	ts = datetime.datetime(2019, 5, 6, 0, 14, 26, 246155)
	return pendulum.parse("2019-05-06 00:14:26.246155Z")


@pytest.mark.parametrize(
	"data",
	[
		"05/06/2019",
		{'year': 2019, 'month': 5, 'day': 6},
		{'year': 2019, 'month': 5, 'day': 6, 'hour': 0, 'minute': 14, 'second': 26, 'microsecond': 246155},
		(2019, 5, 6),
		(2019, 5, 6, 0, 14, 26, 246155),
		"may 6, 2019",
		"06 may 2019",
		pandas.Timestamp(year = 2019, month = 5, day = 6)
	]
)
def test_parse_timestamp_date(data, timestamp):
	result = timetools.Timestamp(data).date()

	assert result == timestamp.date()


@pytest.mark.parametrize("data",
	[
		{'year': 2019, 'month': 5, 'day': 6, 'hour': 0, 'minute': 14, 'second': 26, 'microsecond': 246155},
		(2019, 5, 6, 0, 14, 26, 246155),
		datetime.datetime(2019, 5, 6, 0, 14, 26, 246155),

	]
)
def test_parse_timestamp_datetime(data, timestamp):
	result = timetools.Timestamp(data)
	assert result == timestamp


def test_to_float():
	ts = timetools.Timestamp((2019, 2, 13))
	expected = 2019 + (44 / 365)
	assert float(ts) == expected


def test_to_iso(timestamp):
	expected = timestamp.to_iso8601_string().split('+')[0][:-1]  # Remove the timezone
	result = timetools.Timestamp(timestamp.to_iso8601_string()).to_iso()

	assert result == expected


@pytest.mark.parametrize("value, expected",
	[
		("03/01/20", "2020-03-01")
	]
)
def test_misc(value, expected):
	result = timetools.Timestamp(value)

	assert result.to_iso().split('T')[0] == expected


@pytest.mark.parametrize(
	"string, expected",
	[
		('2016-11-16 22:32:05', datetime.datetime(2016, 11, 16, hour = 22, minute = 32, second = 5)),
		('2010-11-12', datetime.datetime(year = 2010, month = 11, day = 12)),
		('Thu, 31 Mar 2022 22:59:00 -0000', datetime.datetime(year = 2022, month = 3, day = 31)),
		('Thu, 31 Mar 2022 22:59:00 -0000', datetime.datetime(year = 2022, month = 3, day = 31, hour = 0, minute =0, second = 0)),
		("20 Apr. 2022", datetime.datetime(year = 2022, month = 4, day = 20)),
		('13 Sep. 2005', datetime.datetime(year = 2005, month = 9, day = 13)),
		('1 Dec. 2021', datetime.datetime(year = 2021, month = 12, day = 1)),
		('22 Dec. 2019', datetime.datetime(year = 2019, month = 12, day = 22))
	]
)
def test_to_datetime(string, expected):
	# '2016-11-16 22:32:05'
	result = timetools.Timestamp(string).to_datetime()
	assert result == expected

@pytest.mark.parametrize(
	"string, expected",
	[
		("20 Apr. 2022", datetime.datetime(year = 2022, month = 4, day = 20)),
		('13 Sep. 2005', datetime.datetime(year = 2005, month = 9, day = 13)),
		('1 Dec. 2021', datetime.datetime(year = 2021, month = 12, day = 1)),
		('22 Dec. 2019', datetime.datetime(year = 2019, month = 12, day = 22))
	]
)
def test_from_verbal_date(string, expected):

	result = timetools.Timestamp.from_verbal_date(string)

	assert result == expected

@pytest.mark.parametrize(
	"start, stop, expected",
	[

	]
)
def test_timestamp_subtraction(start:str, stop:str, expected: str):
	timestamp_start = timetools.Timestamp(start)
	timestamp_stop = timetools.Timestamp(stop)

	expected_duration = timetools.Duration(expected)

	result = timestamp_stop - timestamp_start

	assert isinstance(result, timetools.Duration)
